@@ -43,23 +43,18 @@ void *applet_reserve_svcctx(struct appctx *appctx, size_t size);
 void applet_reset_svcctx(struct appctx *appctx);
 void appctx_shut(struct appctx *appctx);
 
-struct appctx *appctx_new(struct applet *applet, struct sedesc *sedesc, unsigned long thread_mask);
+struct appctx *appctx_new_on(struct applet *applet, struct sedesc *sedesc, int thr);
 int appctx_finalize_startup(struct appctx *appctx, struct proxy *px, struct buffer *input);
 void appctx_free_on_early_error(struct appctx *appctx);
 
-static inline struct appctx *appctx_new_on(struct applet *applet, struct sedesc *sedesc, uint thr)
-{
-	return appctx_new(applet, sedesc, 1UL << thr);
-}
-
 static inline struct appctx *appctx_new_here(struct applet *applet, struct sedesc *sedesc)
 {
-	return appctx_new(applet, sedesc, tid_bit);
+	return appctx_new_on(applet, sedesc, tid);
 }
 
 static inline struct appctx *appctx_new_anywhere(struct applet *applet, struct sedesc *sedesc)
 {
-	return appctx_new(applet, sedesc, MAX_THREADS_MASK);
+	return appctx_new_on(applet, sedesc, -1);
 }
 
 /* Helper function to call .init applet callback function, if it exists. Returns 0
@@ -71,8 +66,8 @@ static inline int appctx_init(struct appctx *appctx)
 	 * the appctx will be fully initialized. The session and the stream will
 	 * eventually be created. The affinity must be set now !
 	 */
-	BUG_ON((appctx->t->thread_mask & tid_bit) == 0);
-	task_set_affinity(appctx->t, tid_bit);
+	BUG_ON(appctx->t->tid != tid);
+	task_set_thread(appctx->t, tid);
 
 	if (appctx->applet->init)
 		return appctx->applet->init(appctx);
@@ -109,11 +104,11 @@ static inline void appctx_free(struct appctx *appctx)
 	}
 }
 
-/* wakes up an applet when conditions have changed */
-static inline void appctx_wakeup(struct appctx *appctx)
-{
-	task_wakeup(appctx->t, TASK_WOKEN_OTHER);
-}
+/* wakes up an applet when conditions have changed. We're using a macro here in
+ * order to retrieve the caller's place.
+ */
+#define appctx_wakeup(ctx) \
+	_task_wakeup((ctx)->t, TASK_WOKEN_OTHER, MK_CALLER(WAKEUP_TYPE_APPCTX_WAKEUP, 0, 0))
 
 /* returns the stream connector the appctx is attached to, via the sedesc */
 static inline struct stconn *appctx_sc(const struct appctx *appctx)
