@@ -25,7 +25,6 @@
 
 #include <sys/time.h>
 
-#include <import/eb32sctree.h>
 #include <import/eb32tree.h>
 
 #include <haproxy/activity.h>
@@ -87,23 +86,12 @@
 /* tasklets are recognized with nice==-32768 */
 #define TASK_IS_TASKLET(t) ((t)->state & TASK_F_TASKLET)
 
-
 /* a few exported variables */
-extern volatile unsigned long global_tasks_mask; /* Mask of threads with tasks in the global runqueue */
-extern unsigned int grq_total;    /* total number of entries in the global run queue, atomic */
-extern unsigned int niced_tasks;  /* number of niced tasks in the run queue */
-
 extern struct pool_head *pool_head_task;
 extern struct pool_head *pool_head_tasklet;
 extern struct pool_head *pool_head_notification;
 
-#ifdef USE_THREAD
-extern struct eb_root timers;      /* sorted timers tree, global */
-extern struct eb_root rqueue;      /* tree constituting the run queue */
-#endif
-
-__decl_thread(extern HA_SPINLOCK_T rq_lock);  /* spin lock related to run queue */
-__decl_thread(extern HA_RWLOCK_T wq_lock);    /* RW lock related to the wait queue */
+__decl_thread(extern HA_RWLOCK_T wq_lock THREAD_ALIGNED(64));
 
 void __tasklet_wakeup_on(struct tasklet *tl, int thr);
 struct list *__tasklet_wakeup_after(struct list *head, struct tasklet *tl);
@@ -149,9 +137,6 @@ static inline int total_run_queues()
 {
 	int thr, ret = 0;
 
-#ifdef USE_THREAD
-	ret = _HA_ATOMIC_LOAD(&grq_total);
-#endif
 	for (thr = 0; thr < global.nbthread; thr++)
 		ret += _HA_ATOMIC_LOAD(&ha_thread_ctx[thr].rq_total);
 	return ret;
@@ -188,8 +173,8 @@ static inline int task_in_wq(struct task *t)
 /* returns true if the current thread has some work to do */
 static inline int thread_has_tasks(void)
 {
-	return ((int)!!(global_tasks_mask & tid_bit) |
-		(int)!eb_is_empty(&th_ctx->rqueue) |
+	return ((int)!eb_is_empty(&th_ctx->rqueue) |
+	        (int)!eb_is_empty(&th_ctx->rqueue_shared) |
 	        (int)!!th_ctx->tl_class_mask |
 		(int)!MT_LIST_ISEMPTY(&th_ctx->shared_tasklet_list));
 }
@@ -200,20 +185,20 @@ static inline int thread_has_tasks(void)
  * the <file>:<line> from the call place are stored into the task for tracing
  * purposes.
  */
-#define task_wakeup(t, f) _task_wakeup(t, f, __FILE__, __LINE__)
-static inline void _task_wakeup(struct task *t, unsigned int f, const char *file, int line)
+#define task_wakeup(t, f) \
+	_task_wakeup(t, f, MK_CALLER(WAKEUP_TYPE_TASK_WAKEUP, 0, 0))
+
+static inline void _task_wakeup(struct task *t, unsigned int f, const struct ha_caller *caller)
 {
 	unsigned int state;
 
 	state = _HA_ATOMIC_OR_FETCH(&t->state, f);
 	while (!(state & (TASK_RUNNING | TASK_QUEUED))) {
 		if (_HA_ATOMIC_CAS(&t->state, &state, state | TASK_QUEUED)) {
+			caller = HA_ATOMIC_XCHG(&t->caller, caller);
+			BUG_ON((ulong)caller & 1);
 #ifdef DEBUG_TASK
-			if ((unsigned int)t->debug.caller_idx > 1)
-				ABORT_NOW();
-			t->debug.caller_idx = !t->debug.caller_idx;
-			t->debug.caller_file[t->debug.caller_idx] = file;
-			t->debug.caller_line[t->debug.caller_idx] = line;
+			HA_ATOMIC_STORE(&t->debug.prev_caller, caller);
 #endif
 			__task_wakeup(t);
 			break;
@@ -228,8 +213,10 @@ static inline void _task_wakeup(struct task *t, unsigned int f, const char *file
  * that it's possible to unconditionally wakeup the task and drop the RUNNING
  * flag if needed.
  */
-#define task_drop_running(t, f) _task_drop_running(t, f, __FILE__, __LINE__)
-static inline void _task_drop_running(struct task *t, unsigned int f, const char *file, int line)
+#define task_drop_running(t, f) \
+	_task_drop_running(t, f, MK_CALLER(WAKEUP_TYPE_TASK_DROP_RUNNING, 0, 0))
+
+static inline void _task_drop_running(struct task *t, unsigned int f, const struct ha_caller *caller)
 {
 	unsigned int state, new_state;
 
@@ -246,12 +233,10 @@ static inline void _task_drop_running(struct task *t, unsigned int f, const char
 	}
 
 	if ((new_state & ~state) & TASK_QUEUED) {
+		caller = HA_ATOMIC_XCHG(&t->caller, caller);
+		BUG_ON((ulong)caller & 1);
 #ifdef DEBUG_TASK
-		if ((unsigned int)t->debug.caller_idx > 1)
-			ABORT_NOW();
-		t->debug.caller_idx = !t->debug.caller_idx;
-		t->debug.caller_file[t->debug.caller_idx] = file;
-		t->debug.caller_line[t->debug.caller_idx] = line;
+		HA_ATOMIC_STORE(&t->debug.prev_caller, caller);
 #endif
 		__task_wakeup(t);
 	}
@@ -278,8 +263,8 @@ static inline struct task *task_unlink_wq(struct task *t)
 	unsigned long locked;
 
 	if (likely(task_in_wq(t))) {
-		locked = t->state & TASK_SHARED_WQ;
-		BUG_ON(!locked && t->thread_mask != tid_bit && !(global.mode & MODE_STOPPING));
+		locked = t->tid < 0;
+		BUG_ON(t->tid >= 0 && t->tid != tid && !(global.mode & MODE_STOPPING));
 		if (locked)
 			HA_RWLOCK_WRLOCK(TASK_WQ_LOCK, &wq_lock);
 		__task_unlink_wq(t);
@@ -310,71 +295,44 @@ static inline void task_queue(struct task *task)
 		return;
 
 #ifdef USE_THREAD
-	if (task->state & TASK_SHARED_WQ) {
+	if (task->tid < 0) {
 		HA_RWLOCK_WRLOCK(TASK_WQ_LOCK, &wq_lock);
 		if (!task_in_wq(task) || tick_is_lt(task->expire, task->wq.key))
-			__task_queue(task, &timers);
+			__task_queue(task, &tg_ctx->timers);
 		HA_RWLOCK_WRUNLOCK(TASK_WQ_LOCK, &wq_lock);
 	} else
 #endif
 	{
-		BUG_ON(task->thread_mask != tid_bit); // should have TASK_SHARED_WQ
+		BUG_ON(task->tid != tid);
 		if (!task_in_wq(task) || tick_is_lt(task->expire, task->wq.key))
 			__task_queue(task, &th_ctx->timers);
 	}
 }
 
-/* change the thread affinity of a task to <thread_mask>.
+/* Change the thread affinity of a task to <thr>, which may either be a valid
+ * thread number from 0 to nbthread-1, or a negative value to allow the task
+ * to run on any thread.
+ *
  * This may only be done from within the running task itself or during its
  * initialization. It will unqueue and requeue the task from the wait queue
  * if it was in it. This is safe against a concurrent task_queue() call because
  * task_queue() itself will unlink again if needed after taking into account
  * the new thread_mask.
  */
-static inline void task_set_affinity(struct task *t, unsigned long thread_mask)
+static inline void task_set_thread(struct task *t, int thr)
 {
+#ifndef USE_THREAD
+	/* no shared queue without threads */
+	thr = 0;
+#endif
 	if (unlikely(task_in_wq(t))) {
 		task_unlink_wq(t);
-		t->thread_mask = thread_mask;
+		t->tid = thr;
 		task_queue(t);
 	}
-	else
-		t->thread_mask = thread_mask;
-}
-
-/*
- * Unlink the task <t> from the run queue if it's in it. The run queue size and
- * number of niced tasks are updated too. A pointer to the task itself is
- * returned. If the task is in the global run queue, the global run queue's
- * lock will be used during the operation.
- */
-static inline struct task *task_unlink_rq(struct task *t)
-{
-	int is_global = t->state & TASK_GLOBAL;
-	int done = 0;
-
-	if (is_global)
-		HA_SPIN_LOCK(TASK_RQ_LOCK, &rq_lock);
-
-	if (likely(task_in_rq(t))) {
-		eb32sc_delete(&t->rq);
-		done = 1;
+	else {
+		t->tid = thr;
 	}
-
-	if (is_global)
-		HA_SPIN_UNLOCK(TASK_RQ_LOCK, &rq_lock);
-
-	if (done) {
-		if (is_global) {
-			_HA_ATOMIC_AND(&t->state, ~TASK_GLOBAL);
-			_HA_ATOMIC_DEC(&grq_total);
-		}
-		else
-			_HA_ATOMIC_DEC(&th_ctx->rq_total);
-		if (t->nice)
-			_HA_ATOMIC_DEC(&niced_tasks);
-	}
-	return t;
 }
 
 /* schedules tasklet <tl> to run onto thread <thr> or the current thread if
@@ -384,8 +342,10 @@ static inline struct task *task_unlink_rq(struct task *t)
  * <file>:<line> from the call place are stored into the tasklet for tracing
  * purposes.
  */
-#define tasklet_wakeup_on(tl, thr) _tasklet_wakeup_on(tl, thr, __FILE__, __LINE__)
-static inline void _tasklet_wakeup_on(struct tasklet *tl, int thr, const char *file, int line)
+#define tasklet_wakeup_on(tl, thr) \
+	_tasklet_wakeup_on(tl, thr, MK_CALLER(WAKEUP_TYPE_TASKLET_WAKEUP, 0, 0))
+
+static inline void _tasklet_wakeup_on(struct tasklet *tl, int thr, const struct ha_caller *caller)
 {
 	unsigned int state = tl->state;
 
@@ -396,14 +356,12 @@ static inline void _tasklet_wakeup_on(struct tasklet *tl, int thr, const char *f
 	} while (!_HA_ATOMIC_CAS(&tl->state, &state, state | TASK_IN_LIST));
 
 	/* at this point we're the first ones to add this task to the list */
+	caller = HA_ATOMIC_XCHG(&tl->caller, caller);
+	BUG_ON((ulong)caller & 1);
 #ifdef DEBUG_TASK
-	if ((unsigned int)tl->debug.caller_idx > 1)
-		ABORT_NOW();
-	tl->debug.caller_idx = !tl->debug.caller_idx;
-	tl->debug.caller_file[tl->debug.caller_idx] = file;
-	tl->debug.caller_line[tl->debug.caller_idx] = line;
+	HA_ATOMIC_STORE(&tl->debug.prev_caller, caller);
 #endif
-	if (task_profiling_mask & tid_bit)
+	if (_HA_ATOMIC_LOAD(&th_ctx->flags) & TH_FL_TASK_PROFILING)
 		tl->wake_date = now_mono_time();
 	__tasklet_wakeup_on(tl, thr);
 }
@@ -413,7 +371,8 @@ static inline void _tasklet_wakeup_on(struct tasklet *tl, int thr, const char *f
  * DEBUG_TASK is set, the <file>:<line> from the call place are stored into the
  * task for tracing purposes.
  */
-#define tasklet_wakeup(tl) _tasklet_wakeup_on(tl, (tl)->tid, __FILE__, __LINE__)
+#define tasklet_wakeup(tl) \
+	_tasklet_wakeup_on(tl, (tl)->tid, MK_CALLER(WAKEUP_TYPE_TASKLET_WAKEUP, 0, 0))
 
 /* instantly wakes up task <t> on its owner thread even if it's not the current
  * one, bypassing the run queue. The purpose is to be able to avoid contention
@@ -424,12 +383,17 @@ static inline void _tasklet_wakeup_on(struct tasklet *tl, int thr, const char *f
  * TL_URGENT. Great care is taken to be certain it's not queued nor running
  * already.
  */
-#define task_instant_wakeup(t, f) _task_instant_wakeup(t, f, __FILE__, __LINE__)
-static inline void _task_instant_wakeup(struct task *t, unsigned int f, const char *file, int line)
+#define task_instant_wakeup(t, f) \
+	_task_instant_wakeup(t, f, MK_CALLER(WAKEUP_TYPE_TASK_INSTANT_WAKEUP, 0, 0))
+
+static inline void _task_instant_wakeup(struct task *t, unsigned int f, const struct ha_caller *caller)
 {
-	int thr = my_ffsl(t->thread_mask) - 1;
+	int thr = t->tid;
 	unsigned int state;
 
+	if (thr < 0)
+		thr = tid;
+
 	/* first, let's update the task's state with the wakeup condition */
 	state = _HA_ATOMIC_OR_FETCH(&t->state, f);
 
@@ -446,14 +410,12 @@ static inline void _task_instant_wakeup(struct task *t, unsigned int f, const ch
 	BUG_ON_HOT(task_in_rq(t));
 
 	/* at this point we're the first ones to add this task to the list */
+	caller = HA_ATOMIC_XCHG(&t->caller, caller);
+	BUG_ON((ulong)caller & 1);
 #ifdef DEBUG_TASK
-	if ((unsigned int)t->debug.caller_idx > 1)
-		ABORT_NOW();
-	t->debug.caller_idx = !t->debug.caller_idx;
-	t->debug.caller_file[t->debug.caller_idx] = file;
-	t->debug.caller_line[t->debug.caller_idx] = line;
+	HA_ATOMIC_STORE(&t->debug.prev_caller, caller);
 #endif
-	if (task_profiling_mask & tid_bit)
+	if (_HA_ATOMIC_LOAD(&th_ctx->flags) & TH_FL_TASK_PROFILING)
 		t->wake_date = now_mono_time();
 	__tasklet_wakeup_on((struct tasklet *)t, thr);
 }
@@ -467,9 +429,11 @@ static inline void _task_instant_wakeup(struct task *t, unsigned int f, const ch
  * With DEBUG_TASK, the <file>:<line> from the call place are stored into the tasklet
  * for tracing purposes.
  */
-#define tasklet_wakeup_after(head, tl) _tasklet_wakeup_after(head, tl, __FILE__, __LINE__)
+#define tasklet_wakeup_after(head, tl) \
+	_tasklet_wakeup_after(head, tl, MK_CALLER(WAKEUP_TYPE_TASKLET_WAKEUP_AFTER, 0, 0))
+
 static inline struct list *_tasklet_wakeup_after(struct list *head, struct tasklet *tl,
-                                                 const char *file, int line)
+                                                 const struct ha_caller *caller)
 {
 	unsigned int state = tl->state;
 
@@ -480,14 +444,12 @@ static inline struct list *_tasklet_wakeup_after(struct list *head, struct taskl
 	} while (!_HA_ATOMIC_CAS(&tl->state, &state, state | TASK_IN_LIST));
 
 	/* at this point we're the first one to add this task to the list */
+	caller = HA_ATOMIC_XCHG(&tl->caller, caller);
+	BUG_ON((ulong)caller & 1);
 #ifdef DEBUG_TASK
-	if ((unsigned int)tl->debug.caller_idx > 1)
-		ABORT_NOW();
-	tl->debug.caller_idx = !tl->debug.caller_idx;
-	tl->debug.caller_file[tl->debug.caller_idx] = file;
-	tl->debug.caller_line[tl->debug.caller_idx] = line;
+	HA_ATOMIC_STORE(&tl->debug.prev_caller, caller);
 #endif
-	if (task_profiling_mask & tid_bit)
+	if (th_ctx->flags & TH_FL_TASK_PROFILING)
 		tl->wake_date = now_mono_time();
 	return __tasklet_wakeup_after(head, tl);
 }
@@ -496,13 +458,15 @@ static inline struct list *_tasklet_wakeup_after(struct list *head, struct taskl
  * task (or tasklet) wakeup.
  */
 #ifdef DEBUG_TASK
-#define DEBUG_TASK_PRINT_CALLER(t) do {				\
-	printf("%s woken up from %s:%d\n", __FUNCTION__,		\
-	       (t)->debug.caller_file[(t)->debug.caller_idx],	\
-	       (t)->debug.caller_line[(t)->debug.caller_idx]);	\
+#define DEBUG_TASK_PRINT_CALLER(t) do {						\
+	const struct ha_caller *__caller = (t)->caller;			\
+	printf("%s woken up from %s(%s:%d)\n", __FUNCTION__,		\
+	       __caller ? __caller->func : NULL, \
+	       __caller ? __caller->file : NULL, \
+	       __caller ? __caller->line : 0); \
 } while (0)
 #else
-#define DEBUG_TASK_PRINT_CALLER(t)
+#define DEBUG_TASK_PRINT_CALLER(t) do { } while (0)
 #endif
 
 
@@ -523,26 +487,23 @@ static inline void tasklet_remove_from_tasklet_list(struct tasklet *t)
 /*
  * Initialize a new task. The bare minimum is performed (queue pointers and
  * state).  The task is returned. This function should not be used outside of
- * task_new(). If the thread mask contains more than one thread, TASK_SHARED_WQ
- * is set.
+ * task_new(). If the thread ID is < 0, the task may run on any thread.
  */
-static inline struct task *task_init(struct task *t, unsigned long thread_mask)
+static inline struct task *task_init(struct task *t, int tid)
 {
 	t->wq.node.leaf_p = NULL;
 	t->rq.node.leaf_p = NULL;
 	t->state = TASK_SLEEPING;
-	t->thread_mask = thread_mask;
-	if (atleast2(thread_mask))
-		t->state |= TASK_SHARED_WQ;
+#ifndef USE_THREAD
+	/* no shared wq without threads */
+	tid = 0;
+#endif
+	t->tid = tid;
 	t->nice = 0;
 	t->calls = 0;
 	t->wake_date = 0;
-	t->cpu_time = 0;
-	t->lat_time = 0;
 	t->expire = TICK_ETERNITY;
-#ifdef DEBUG_TASK
-	t->debug.caller_idx = 0;
-#endif
+	t->caller = NULL;
 	return t;
 }
 
@@ -557,9 +518,7 @@ static inline void tasklet_init(struct tasklet *t)
 	t->process = NULL;
 	t->tid = -1;
 	t->wake_date = 0;
-#ifdef DEBUG_TASK
-	t->debug.caller_idx = 0;
-#endif
+	t->caller = NULL;
 	LIST_INIT(&t->list);
 }
 
@@ -577,38 +536,29 @@ static inline struct tasklet *tasklet_new(void)
 }
 
 /*
- * Allocate and initialise a new task. The new task is returned, or NULL in
- * case of lack of memory. The task count is incremented. This API might change
- * in the near future, so prefer one of the task_new_*() wrappers below which
- * are usually more suitable. Tasks must be freed using task_free().
+ * Allocate and initialize a new task, to run on global thread <thr>, or any
+ * thread if negative. The task count is incremented. The new task is returned,
+ * or NULL in case of lack of memory. It's up to the caller to pass a valid
+ * thread number (in tid space, 0 to nbthread-1, or <0 for any). Tasks created
+ * this way must be freed using task_destroy().
  */
-static inline struct task *task_new(unsigned long thread_mask)
+static inline struct task *task_new_on(int thr)
 {
 	struct task *t = pool_alloc(pool_head_task);
 	if (t) {
 		th_ctx->nb_tasks++;
-		task_init(t, thread_mask);
+		task_init(t, thr);
 	}
 	return t;
 }
 
-/* Allocate and initialize a new task, to run on global thread <thr>. The new
- * task is returned, or NULL in case of lack of memory. It's up to the caller
- * to pass a valid thread number (in tid space, 0 to nbthread-1). The task
- * count is incremented.
- */
-static inline struct task *task_new_on(uint thr)
-{
-	return task_new(1UL << thr);
-}
-
 /* Allocate and initialize a new task, to run on the calling thread. The new
  * task is returned, or NULL in case of lack of memory. The task count is
  * incremented.
  */
 static inline struct task *task_new_here()
 {
-	return task_new(tid_bit);
+	return task_new_on(tid);
 }
 
 /* Allocate and initialize a new task, to run on any thread. The new task is
@@ -616,7 +566,7 @@ static inline struct task *task_new_here()
  */
 static inline struct task *task_new_anywhere()
 {
-	return task_new(MAX_THREADS_MASK);
+	return task_new_on(-1);
 }
 
 /*
@@ -631,11 +581,11 @@ static inline void __task_free(struct task *t)
 	}
 	BUG_ON(task_in_wq(t) || task_in_rq(t));
 
+	BUG_ON((ulong)t->caller & 1);
 #ifdef DEBUG_TASK
-	if ((unsigned int)t->debug.caller_idx > 1)
-		ABORT_NOW();
-	t->debug.caller_idx |= 2; // keep parity and make sure to crash if used after free
+	HA_ATOMIC_STORE(&t->debug.prev_caller, HA_ATOMIC_LOAD(&t->caller));
 #endif
+	HA_ATOMIC_STORE(&t->caller, (void*)1); // make sure to crash if used after free
 
 	pool_free(pool_head_task, t);
 	th_ctx->nb_tasks--;
@@ -674,11 +624,11 @@ static inline void tasklet_free(struct tasklet *tl)
 	if (MT_LIST_DELETE(list_to_mt_list(&tl->list)))
 		_HA_ATOMIC_DEC(&ha_thread_ctx[tl->tid >= 0 ? tl->tid : tid].rq_total);
 
+	BUG_ON((ulong)tl->caller & 1);
 #ifdef DEBUG_TASK
-	if ((unsigned int)tl->debug.caller_idx > 1)
-		ABORT_NOW();
-	tl->debug.caller_idx |= 2; // keep parity and make sure to crash if used after free
+	HA_ATOMIC_STORE(&tl->debug.prev_caller, HA_ATOMIC_LOAD(&tl->caller));
 #endif
+	HA_ATOMIC_STORE(&tl->caller, (void*)1); // make sure to crash if used after free
 	pool_free(pool_head_tasklet, tl);
 	if (unlikely(stopping))
 		pool_flush(pool_head_tasklet);
@@ -704,7 +654,7 @@ static inline void task_schedule(struct task *task, int when)
 		return;
 
 #ifdef USE_THREAD
-	if (task->state & TASK_SHARED_WQ) {
+	if (task->tid < 0) {
 		/* FIXME: is it really needed to lock the WQ during the check ? */
 		HA_RWLOCK_WRLOCK(TASK_WQ_LOCK, &wq_lock);
 		if (task_in_wq(task))
@@ -712,12 +662,12 @@ static inline void task_schedule(struct task *task, int when)
 
 		task->expire = when;
 		if (!task_in_wq(task) || tick_is_lt(task->expire, task->wq.key))
-			__task_queue(task, &timers);
+			__task_queue(task, &tg_ctx->timers);
 		HA_RWLOCK_WRUNLOCK(TASK_WQ_LOCK, &wq_lock);
 	} else
 #endif
 	{
-		BUG_ON((task->thread_mask & tid_bit) == 0); // should have TASK_SHARED_WQ
+		BUG_ON(task->tid != tid);
 		if (task_in_wq(task))
 			when = tick_first(when, task->expire);
 
@@ -727,6 +677,22 @@ static inline void task_schedule(struct task *task, int when)
 	}
 }
 
+/* returns the string corresponding to a task type as found in the task caller
+ * locations.
+ */
+static inline const char *task_wakeup_type_str(uint t)
+{
+	switch (t) {
+	case WAKEUP_TYPE_TASK_WAKEUP          : return "task_wakeup";
+	case WAKEUP_TYPE_TASK_INSTANT_WAKEUP  : return "task_instant_wakeup";
+	case WAKEUP_TYPE_TASKLET_WAKEUP       : return "tasklet_wakeup";
+	case WAKEUP_TYPE_TASKLET_WAKEUP_AFTER : return "tasklet_wakeup_after";
+	case WAKEUP_TYPE_TASK_DROP_RUNNING    : return "task_drop_running";
+	case WAKEUP_TYPE_APPCTX_WAKEUP        : return "appctx_wakeup";
+	default                               : return "?";
+	}
+}
+
 /* This function register a new signal. "lua" is the current lua
  * execution context. It contains a pointer to the associated task.
  * "link" is a list head attached to an other task that must be wake
