@@ -33,13 +33,13 @@
 #ifdef  CONFIG_PRODUCT_BRANCH
 #define PRODUCT_BRANCH    CONFIG_PRODUCT_BRANCH
 #else
-#define PRODUCT_BRANCH   "2.6"
+#define PRODUCT_BRANCH   "2.7"
 #endif
 
 #ifdef  CONFIG_PRODUCT_STATUS
 #define PRODUCT_STATUS    CONFIG_PRODUCT_STATUS
 #else
-#define PRODUCT_STATUS   "Status: long-term supported branch - will stop receiving fixes around Q2 2027."
+#define PRODUCT_STATUS   "Status: stable branch - will stop receiving fixes around Q1 2024."
 #endif
 
 #ifdef CONFIG_PRODUCT_URL_BUGS
