@@ -349,7 +349,7 @@ int httpclient_req_gen(struct httpclient *hc, const struct ist url, enum http_me
 	if (!htx_add_endof(htx, HTX_BLK_EOH))
 		goto error;
 
-	if (isttest(payload)) {
+	if (isttest(payload) && istlen(payload)) {
 		/* add the payload if it can feat in the buffer, no need to set
 		 * the Content-Length, the data will be sent chunked */
 		if (!htx_add_data_atonce(htx, payload))
